from transformers import BertTokenizer, GPT2Tokenizer, RobertaTokenizer, DistilBertTokenizer, AlbertTokenizer, ElectraTokenizer
import numpy as np
import tensorflow as tf

class Text2Number():
    def __init__(self, batch_size=32):
        self.token = None
        self.batch_size = batch_size

    def tokenize_bert(self, data, seq_len=128, model="bert-base-cased"):
        '''
            Computationally expensive due to their large size, and the tokenization process can be slower
        '''
        tokenizer = BertTokenizer.from_pretrained(model)
        token = tokenizer.encode_plus(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_attention_mask=True,)
        self.token = token

    def tokenize_gpt2(self, data, seq_len=128, model="gpt2-medium"):
        '''
            Unidirectional
        '''
        tokenizer = GPT2Tokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_robert(self, data, seq_len=128, model="roberta-base"):
        '''
            Requiring more computational resources for training and inference.
        '''
        tokenizer = RobertaTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_distilBERT(self, data, seq_len=128, model="distilbert-base-uncased"):
        '''
            Scenarios where memory and computational resources are limited
        '''
        tokenizer = DistilBertTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_Albert(self, data, seq_len=128, model="albert-base-v2"):
        '''
            Fewer parameters, making them more memory-efficient and faster during training and inference
        '''
        tokenizer = AlbertTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_attention_mask=True,)
        self.token = token

    def tokenize_Electra(self, data, seq_len=128, model="google/electra-base-discriminator"):
        '''
            Rely on a discriminator-generator setup, which means they may not capture certain aspects of language semantics
        '''
        tokenizer = ElectraTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_attention_mask=True,)
        self.token = token

    def data_tokenized(self, data, labels, seq_len=128):
        '''
            Creates the train and validation datasets
        '''
        train_ds_ids = []
        train_ds_mask = []
        train_label = []
        labels = np.array(labels.tolist()).reshape(-1,1)

        tokenizer = BertTokenizer.from_pretrained("bert-base-cased")
        for i, val in enumerate(data.tolist()):
            val = tokenizer.encode_plus(val,
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_attention_mask=True,)

            train_ds_ids.append(val['input_ids'])
            train_ds_mask.append(val['attention_mask'])
            train_label.append(labels[i])

        train_ds_ids = tf.convert_to_tensor(train_ds_ids)
        train_ds_mask = tf.convert_to_tensor(train_ds_mask)
        train_label = tf.convert_to_tensor(train_label)

        return train_ds_ids,train_ds_mask, train_label
    
    def map_func_test(self, input_ids, masks):
        return {'input_ids':input_ids,
                'attention_mask':masks}
    
    def map_func(self, input_ids, masks, labels):
        return {'input_ids':input_ids,
                'attention_mask':masks}, labels
    
    def get(self):
        return self.token
    


from ...model_data.model_data import ModelData
from sklearn.model_selection import train_test_split, KFold


def Complexity_Level0(model, X, y):
    print(f"Phase I \t Phase II \t Phase III \t Phase IV\n")
    importance_columns = {}
    for col in X.columns:
        importance_columns[col] = 0

    ## fix null columns
    model_data = ModelData(X)
    columns_null = model_data.find_null_column()
    for col in columns_null:
        column_data_type = X[col].dtype
        if column_data_type == "object":
            model_data.fix_column_str(col)
        else:
            model_data.fix_columns([col], "mean")
    print(f"Completed", end=" ")
    
    ## Text column to dummies
    for col in X.columns:
        print(col)
        if X[col].dtype == "object" and X[col].nunique() < 10:
            model_data.dummies([col])
    X = model_data.return_dataframe()

    ## Train model in diffrent samples to extract importance of columns
    kfold = KFold(n_splits=5, shuffle=True, random_state=42)
    for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
        X_train, y_train = X.iloc[train_index], y.iloc[train_index]
        model.create_optuna(X_train,y_train,n_trials=50)
        feature_model = model.get()
        features = feature_model.get_feature_importance()
        for i,col in enumerate(X.columns):
            importance_columns[col] += features[i]
    print("-------Complete", end=" ")
    
    ## Acceptting column that satify the requirenments 
    mean_importance = sum(importance_columns.values()) / len(importance_columns)
    threshold_importance = 0.75 * mean_importance
    columns = []
    for k,v in importance_columns.items():
        if v > threshold_importance:
            columns.append(k)
    print("-------Complete", end=" ")

    X = X[columns]
    model.create_optuna(X,y,n_trials=300)
    print("-------Complete", end=" ")
    return model.get(), columns




from nltk.stem import WordNetLemmatizer
from nltk import pos_tag
from nltk.stem.porter import PorterStemmer
from nltk.corpus import stopwords, wordnet
from collections import Counter
from spellchecker import SpellChecker
import nltk
import pandas as pd
import emoji
import string
import re


class TextPreprocessing():
    def __init__(self, text:pd.DataFrame):
        self.text = text

    def lowercase(self):
        self.text = self.text.apply(lambda x: lower(x))

    def punctuation(self):
        self.text = self.text.apply(lambda x: remove_punctuation(x))

    def stopwords(self, language='english'):
        try:
            nltk.data.find('corpora/stopwords')
        except LookupError:
            nltk.download('stopwords')
        
        stopword = set(stopwords.words(language))
        self.text = self.text.apply(lambda x: remove_stopwords(x, stopword))

    def frequent_words(self, most_common=10):
        words = Counter()
        for text in self.text:
            for word in text.split():
                words[word] += 1
        return words.most_common(most_common)
    
    def frequent_words(self, number = 0, words=None, auto=False):
        if auto:
            frequent_words = self.frequent_words(number)
            frequent_words = set(word for  (word, _) in frequent_words)
        else:
            frequent_words = words
        self.text = self.text.apply(lambda x: remove_frequent_words(x, frequent_words))

    def special_characters(self, change_into=" ",syntax='[^a-zA-Z0-9]'):
        self.text = self.text.apply(lambda x: remove_special_characters(x, change_into, syntax))

    def stemming(self):
        ps = PorterStemmer()
        self.text = self.text.apply(lambda x: stemming(x, ps))

    def lemmatization(self):
        lemmat = WordNetLemmatizer()
        wordnet_map = {
            "N": wordnet.NOUN,
            "V": wordnet.VERB,
            "J": wordnet.ADJ,
            "R": wordnet.ADV,
        }
        self.text = self.text.apply(lambda x: lemat_word(x, lemmat, wordnet_map))

    def url(self, change_to=""):
        self.text = self.text.apply(lambda x: remove_url(x, change_to))

    def html(self):
        self.text = self.text.apply(lambda x: remove_html(x))

    def spelling(self):
        spell = SpellChecker()
        self.text = self.text.apply(lambda x: correct_spelling(x, spell))

    def emoji(self, lang='en'):
        self.text = self.text.apply(lambda x: strip_emoji(x, lang))

    def get(self):
        return self.text


def lower(text):
    return text.lower()

def strip_emoji(text, lang):
    return emoji.demojize(text, language=lang)

def remove_punctuation(text):
    punctuations = string.punctuation
    return text.translate(str.maketrans(' ', ' ', punctuations))
    
def remove_stopwords(text, stopwords):
    return " ".join([word for word in text.split() if word not in stopwords])

def remove_frequent_words(text, remove_words):
    return " ".join([word for word in text.split() if word not in remove_words])

def remove_special_characters(text, change_into, syntax):
    text = re.sub(syntax,change_into, text)
    text = re.sub('\s+', ' ', text)
    return text

def stemming(text, stemmer):
    return " ".join([stemmer.stem(word) for word in text.split()])

def lemat_word(text, lemat, wordnet_map):
    pos_text = pos_tag(text.split())
    return " ".join([lemat.lemmatize(word, wordnet_map.get(pos[0], wordnet.NOUN)) for word, pos in pos_text])

def remove_url(text, change_to):
    return re.sub(r'https?://\S+|www\.\S+', change_to, text)

def remove_html(text):
    return re.sub(r'<.*?>', '', text)

def correct_spelling(text, spell):
    corrected_text = []
    mispelled = spell.unknown(text.split())
    for word in text.split():
        if word in mispelled:
            corrected_word = spell.correction(word)
            if corrected_word is not None:
                corrected_text.append(corrected_word)
            else:
                corrected_text.append(word)
        else:
            corrected_text.append(word)
    return " ".join(corrected_text)



from sklearn.model_selection import GridSearchCV, train_test_split,KFold
from sklearn.metrics import mean_squared_error,r2_score
from ..metrics_regression import Metrics
import optuna.visualization as vis
import lightgbm as lgb
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class LGB(Metrics):
    def __init__(self):
        self.model = None
        self.parameters = None
    
    def put(self, model):
        self.model = model

    def create(self, X, y, params=None):
        if params == None:
            gbm = lgb.LGBMRegressor()
            gbm.fit(X,y)
            self.model = gbm
        else:
            gbm = lgb.LGBMRegressor(**params)
            gbm.fit(X,y)
            self.model = gbm
            self.parameters = params

    def create_grid(self, X, y, params=None, cv=3):
        params_columns = ['boosting_type', 'objective', 'num_leaves',
                          'n_estimators', 'max_depth', 'min_child_samples',
                          'subsample', 'colsample_bytree', 'seed', 'max_bin']
        params_basic = {
            'boosting_type': ['gbdt'],
            'objective': ['regression'],
            'num_leaves': [20,200],
            'n_estimators': [30, 100],
            'max_depth': [5, 32],
            'min_child_samples': [30,70],
            'subsample': [0.3, 0.9],
            'colsample_bytree': [0.5, 1.0],
            'seed': [42],
            'max_bin': [255]
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        gbm = lgb.LGBMRegressor()
        grid_search = GridSearchCV(gbm, params, cv=cv, n_jobs=-1)
        grid_search.fit(X, y)
        best_model = grid_search.best_estimator_
        self.model = best_model
        self.parameters = grid_search.best_params_

    def create_optuna(self, X, y, params=None, n_trials=5, show_plot=False, show_features=False):
        params_columns = ['boosting_type', 'objective', 'num_leaves',
                          'n_estimators', 'max_depth', 'min_child_samples',
                          'subsample', 'colsample_bytree', 'seed', 'max_bin']
        params_basic = {
            'boosting_type': ['gbdt'],
            'objective': 'regression',
            'num_leaves': [10,100],
            'n_estimators': [10, 200],
            'max_depth': [1, 6],
            'min_child_samples': [10,70],
            'subsample': 0.9,
            'colsample_bytree': 0.7,
            'seed': 42,
            'max_bin': 255
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
        def objective(trial):
            param = {
                'boosting_type': trial.suggest_categorical("boosting_type", params['boosting_type']),
                'objective': trial.suggest_categorical("objective", params['objective']),
                'num_leaves': trial.suggest_int('num_leaves', params['num_leaves'][0], params['num_leaves'][1]),
                'n_estimators': trial.suggest_int('n_estimators', params['n_estimators'][0], params['n_estimators'][1]),
                'max_depth': trial.suggest_int("max_depth", params['max_depth'][0], params['max_depth'][1]),
                'min_child_samples': trial.suggest_int("min_child_samples", params['min_child_samples'][0], params['min_child_samples'][1]),
                'subsample': trial.suggest_float("subsample", params['subsample']),
                'colsample_bytree': trial.suggest_float("colsamlpe_bytree", params['colsample_bytree']),
                'seed': trial.suggest_int("seed", params['seed'], params['seed']),
                'max_bin': trial.suggest_int("max_bin", params['max_bin'], params['max_bin'])
            }
            gbm = lgb.LGBMRegressor(**param)
            gbm.fit(X_train, y_train)
            y_pred = gbm.predict(X_test)
            mse = mean_squared_error(y_test, y_pred)
            return mse
        study = optuna.create_study(direction='minimize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()

        best_params = study.best_params
        gbm_best = lgb.LGBMRegressor(**best_params)
        gbm_best.fit(X, y)
        self.model= gbm_best
        self.parameters = best_params

    def score(self, X, y):
        preds = np.round(self.model.predict(X))
        return self.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],))
        r2 = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += self.predict(df_test)/n_splits
            val_pred = self.predict(X_test)
            r2.append(r2_score(y_test,val_pred))

            print(f"{i} Fold scored: {r2[i]}")

        print(f"Mean r2_score {np.mean(r2)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path='lgb.txt'):
        self.model.save_model(model_path)

    def load(self, model_path):
        self.model =lgb.Booster(model_file=model_path)


from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score, median_absolute_error
import numpy as np

class Metrics():
  def __init__(self):
    self.mae = 0
    self.mse = 0
    self.rmse = 0 
    self.r2 = 0


  def calculate_metrics(self, y_true, y_preds):
    self.mae = mean_absolute_error(y_true, y_preds)
    self.mse= mean_squared_error(y_true, y_preds)
    self.rmse = np.sqrt(self.mse)
    self.r2 = r2_score(y_true, y_preds)

    metrics = {"mae": self.mae,
               "mse": self.mse,
               "rmse": self.rmse,
               "r2": self.r2}
    
    return metrics


from .cat_automated import CatAutomated
from .gradient_boosting_automated import GradientBoostingAutomated
from .lgb_automated import LGBAutomated
from .xgb_automated import XGBAutomated
from ..algorithms.complexity_level0 import Complexity_Level0
from ..algorithms.complexity_level1 import Complexity_Level1


import tensorflow as tf
from sklearn.metrics import log_loss
import optuna.visualization as vis
from ...classification.metrics import Metrics
from transformers import TFAutoModel
import optuna
import numpy as np
optuna.logging.disable_default_handler()


class TransformerModel():
    def __init__(self, num_classes, seq_len=128, bert_model_name='bert-base-uncased'):
        self.classes = num_classes
        self.bert_model = TFAutoModel.from_pretrained(bert_model_name)
        self.layers = []
        self.seq_len = seq_len
        self.model = None

    def add_dense(self, dimensions=128, activation="relu"):
        self.layers.append(tf.keras.layers.Dense(dimensions, activation=activation))

    def build(self):
        input_ids = tf.keras.Input(shape=(self.seq_len, ),name="input_ids", dtype=tf.int32)
        attention_mask = tf.keras.Input(shape=(self.seq_len, ),name="attention_mask", dtype=tf.int32)

        bert = self.bert_model([input_ids, attention_mask])
        x = bert[1]
        for layer in self.layers:
            x = layer(x)
        if self.classes < 2:
            y = tf.keras.layers.Dense(self.classes, activation='sigmoid')(x)
        else:
            y = tf.keras.layers.Dense(self.classes, activation='softmax')(x)

        self.model = tf.keras.Model(inputs=[input_ids, attention_mask], outputs=y)

    def compile(self, loss, optimizer, metrics):
        self.build()
        self.model.compile(loss=loss,
                          optimizer=optimizer,
                          metrics=metrics)
    
    def fit(self, X_data,X_label, val_data, val_label, epochs=1, batch=32, layer_untrained=-1):
        if layer_untrained >= 0:
            self.model.layers[layer_untrained].trainable = False
        self.model.fit(X_data, X_label,
                       validation_data=(val_data, val_label),
                       epochs=epochs)
        
#     def fit_optuna(self, X_data, X_label, val_data, val_label, batch=32, layer_untrained=-1, params=None, show_plot=False, show_features=False):
#         params_columns = ["layers","layers_dimensions",]
#         params_basic = {
#             "layers": [2, 10],
#             "layers_dimensions": [32, 1024],
#         }
#         if params is None:
#             params = params_basic
#         else:
#             for parameter in params_columns:
#                 if parameter not in params.keys():
#                     params[parameter] = params_basic[parameter]
        
#         def objective(trial):
#             self.reset()
#             layers = trial.suggest_int("Layers", params["layers"][0], params["layers"][1])
#             for i in range(layers):
#                 self.add_dense(trial.suggest_int(f'Layer_{i}', params["layers_dimensions"][0], params["layers_dimensions"][1]))
#             self.fit(X_data, X_label, val_data, val_label,batch,layer_untrained)
#             preds = self.predict(val_data)
#             acc = accuracy_score(val_label, preds)
#             return acc
        
#         study = optuna.create_study(direction="maximize",pruner=optuna.pruners.MedianPruner())
#         study.optimize(objective, n_trials=n_trials)
#         if show_plot:
#             optimization_history_plot = vis.plot_optimization_history(study)
#             optimization_history_plot.show()
#         if show_features:
#             param_importance_plot = vis.plot_param_importances(study)
#             param_importance_plot.show()
            
#         best_params = study.best_trial.params
#         self.reset()
#         for i in range(best_params["Layers"]):
#             self.add_dense(best_params[f"Layer_{i}"])
#         self.fit(X_data, X_label, val_data, val_label,batch,layer_untrained)
                

    def predict(self, data):
        return self.model.predict(data)
    
    def summary(self):
        self.model.summary()
        
    def reset(self):
        self.layers = []





    


import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay

def plot_history(history):
    '''
    Takes history of our model and plot accuracy and loss function

    parms:
    history - model history

    output:
    -
    '''
    loss = history.history["loss"]
    acc = history.history["accuracy"]
    number_of_elements = len(loss)
    epochs = np.linspace(1,number_of_elements,number_of_elements)

    ## Plotting functions
    fig,ax = plt.subplots(2,1, figsize=(10,7))
    ax[0].plot(epochs, acc)
    ax[0].set_title("Accuracy")  
    ax[1].plot(epochs, loss)
    ax[1].set_title("Loss function")
    plt.show()

def show_confusion_matrix(model,X,y_true):
    pred = model.predict(X)
    confusion_matrix = confusion_matrix(y_true, pred)
    cm_display = ConfusionMatrixDisplay(confusion_matrix = confusion_matrix)
    cm_display.plot()
    plt.show()


from sklearn.model_selection import GridSearchCV, train_test_split,KFold
from sklearn.metrics import mean_squared_error,r2_score
from ..metrics_regression import Metrics
from sklearn.linear_model import SGDRegressor
from joblib import dump, load
import numpy as np

class SGD(Metrics):
    def __init__(self):
        self.model = None
        self.parameters = None

    def put(self, model):
        self.model = model

    def create(self, X, y, params=None):
        if params == None:
            gbm = SGDRegressor()
            gbm.fit(X,y)
            self.model = gbm
        else:
            gbm = SGDRegressor(**params)
            gbm.fit(X,y)
            self.model = gbm
            self.parameters = params
         
    def create_grid(self,X,y, params=None, cv=3):
        params_columns = ['loss', 'penalty', 'alpha', 'learning_rate', 'eta0', 'random_state']

        params_basic = {
            'loss': ['squared_epsilon_insensitive', 'epsilon_insensitive', 'squared_error', 'huber'],
            'penalty': ['l1', 'l2', 'elasticnet'],
            'alpha': [0.0001, 0.001, 0.01],
            'learning_rate': ['constant', 'optimal', 'invscaling', 'adaptive'],
            'eta0': [0.001, 0.01, 0.1],
            'random_state': [42]
        }
        if params == None:
                params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]
        
        sgd = SGDRegressor()
        grid_search = GridSearchCV(sgd, params, cv=cv)
        grid_search.fit(X, y)
        best_params = grid_search.best_params_
        best_model = grid_search.best_estimator_
        self.model = best_model
        self.parameters = best_params

    def score(self, X, y):
        preds = np.round(self.model.predict(X))
        return self.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],))
        r2 = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += self.predict(df_test)/n_splits
            val_pred = self.predict(X_test)
            r2.append(r2_score(y_test,val_pred))

            print(f"{i} Fold scored: {r2[i]}")

        print(f"Mean r2_score {np.mean(r2)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path="sgd.joblib"):
        dump(self.model, model_path)
    
    def load(self, model_path):
        self.model = load(model_path)


from sklearn.model_selection import train_test_split, GridSearchCV, KFold
from sklearn.ensemble import GradientBoostingClassifier
from ..metrics import Metrics
import optuna.visualization as vis
from sklearn.metrics import accuracy_score, roc_auc_score
from joblib import dump, load
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class GradientBoosting(Metrics):
    def __init__(self):
        self.metric = Metrics()
        self.model = None
        self.parameters = None

    def create(self, X, y, params=None):
        if params == None:
            cat = GradientBoostingClassifier()
            cat.fit(X,y)
            self.model = cat
        else:
            cat = GradientBoostingClassifier(**params)
            cat.fit(X,y)
            self.model = cat
            self.parameters = params

    def create_grid(self, X, y, params=None, cv=2):
        params_columns = ["loss","learning_rate", "n_estimators",
                          "subsample","min_samples_split","max_depth",
                          "random_state","max_features","verbose",
                          "validation_fraction"]
        params_basic = {
                'loss': ['log_loss'],
                'learning_rate': [0.001,0.01,0.1],
                'n_estimators': [10, 100],
                'subsample': [0.8],
                'min_samples_split': [3,10],
                'max_depth': [3, 10],
                'random_state': [42],
                'max_features': ['sqrt', None],
                'verbose': [0],
                'validation_fraction': [0.2] 
            }
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]
            
        grad = GradientBoostingClassifier()
        grid_search = GridSearchCV(grad,params,cv=cv)
        grid_search.fit(X,y)
        self.model = grid_search.best_estimator_
        self.parameters = grid_search.best_params_


    def create_optuna(self, X, y, params=None, n_trials=3, show_plot=False, show_features=False):
        params_columns = ["loss","learning_rate", "n_estimators",
                          "subsample","min_samples_split","max_depth",
                          "random_state","max_features","verbose",
                          "validation_fraction"]
        params_basic = {
                'loss': ['log_loss'],
                'learning_rate': [0.001,0.1],
                'n_estimators': [10, 500],
                'subsample': 0.8,
                'min_samples_split': [3,10],
                'max_depth': [1, 8],
                'random_state': 42,
                'max_features': ['sqrt'],
                'verbose': 0,
                'validation_fraction': 0.2
            }
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X,y, test_size=0.2, random_state=42)
        def objective(trail):
            param = {
                'loss': trail.suggest_categorical("loss", params['loss']),
                'learning_rate': trail.suggest_float("learning_rate", params['learning_rate'][0], params['learning_rate'][1]),
                'n_estimators': trail.suggest_int('n_estimators', params['n_estimators'][0], params['n_estimators'][1]),
                'subsample': trail.suggest_float("subsample", params['subsample'], params['subsample']),
                'min_samples_split': trail.suggest_int('min_samples_split', params['min_samples_split'][0], params['min_samples_split'][1]),
                'max_depth': trail.suggest_int('max_depth', params['max_depth'][0], params['max_depth'][1]),
                'random_state': trail.suggest_int("random_state", params["random_state"], params["random_state"]),
                'max_features': trail.suggest_categorical("max_features", params['max_features']),
                'verbose': trail.suggest_int("verbose", params["verbose"], params["verbose"]),
                'validation_fraction': trail.suggest_float('validation_fraction', params['validation_fraction'], params['validation_fraction']) 
            }
            gb = GradientBoostingClassifier(**param)
            gb.fit(X_train, y_train)
            preds = gb.predict(X_test)
            accuracy = accuracy_score(y_test, preds)
            return accuracy

        study = optuna.create_study(direction='maximize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()
            
        best_params = study.best_params
        gb_best = GradientBoostingClassifier(**best_params)
        gb_best.fit(X, y)
        self.model= gb_best
        self.parameters = best_params


    def score(self, X, y):
        preds = self.model.predict(X)
        return self.metric.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None, classes=1):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],classes))
        roc = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            predictions += np.expand_dims(self.predict(df_test)/n_splits, axis=-1)

            val_pred = self.predict(X_test)
            if classes > 1:
              roc.append(roc_auc_score(y_test,val_pred,multi_class='ovr'))
            else:
              roc.append(roc_auc_score(y_test,val_pred))
            print(f"{i} Fold scored: {roc[i]}")

        print(f"Mean roc score {np.mean(roc)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path="gradient_boosting.joblib"):
        dump(self.model, model_path)
    
    def load(self, model_path):
        self.model = load(model_path)